# Copyright (C) 2022 H. Turgut Uyar <uyar@tekir.org>
#
# Kırlent is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Kırlent is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Kırlent.  If not, see <http://www.gnu.org/licenses/>.

__version__ = "0.3"

import json
from pathlib import Path

from invoke import Collection, Program

from . import tasks


config: Path = Path("kirlent.json")
if not config.exists():
    config = Path(__file__).parent / "kirlent.json"

namespace: Collection = Collection.from_module(tasks)
namespace.configure(json.loads(config.read_text()))

program: Program = Program(namespace=namespace, version=__version__)



# Copyright (C) 2022 H. Turgut Uyar <uyar@tekir.org>
#
# Kırlent is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Kırlent is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Kırlent.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path

from invoke import Collection, task

from . import pdf, slides
from .utils import COPY


@task
def setup(c):
    c.run("python -m pip --require-virtualenv install -U kirlent_docutils")
    c.run("npm install decktape")

    c.run(COPY % {
        "src": Path(__file__).parent / "env",
        "dst": Path(".env"),
    })

    c.run(COPY % {
        "src": Path(__file__).parent / "kirlent.json",
        "dst": Path("kirlent.json"),
    })


namespace: Collection = Collection(setup, slides, pdf)



# Copyright (C) 2022 H. Turgut Uyar <uyar@tekir.org>
#
# Kırlent is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Kırlent is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Kırlent.  If not, see <http://www.gnu.org/licenses/>.

from io import StringIO
from pathlib import Path
from typing import Dict

from invoke import Context, task

from .utils import MKDIR, collect_assets, relative_path, up_to_date


BUILDERS: Dict[str, str] = {
    "simple": "kirlent2slides",
    "impressjs": "kirlent2impressjs",
    "revealjs": "kirlent2revealjs",
}

BUILD_SLIDES: str = "%(builder)s %(options)s %(in)s"


def slides(c: Context, src: Path, output: Path, *, framework: str) -> None:
    suffixes = "".join(src.suffixes)
    dst_name = src.name.replace(suffixes, f"-{framework}{suffixes}")
    dst = Path(output, dst_name).with_suffix(".html")
    if not up_to_date(dst, [src]):
        if not dst.parent.exists():
            c.run(MKDIR % {"dir": relative_path(dst.parent)})
        lang = src.name.split(".")[1]
        options = c.slides.options.base + c.slides.options.get(framework, [])
        cli_options = " ".join(options) % {
            "lang": lang,
            "size": c.slides.size,
        }
        out = StringIO()
        c.run(
            BUILD_SLIDES % {
                "builder": BUILDERS[framework],
                "options": cli_options,
                "in": relative_path(src),
            },
            out_stream=out,
        )
        text = collect_assets(c, out.getvalue(), base=output, wrt=src.parent)
        dst.write_text(text)
    c.config[f"{framework}:output"] = str(dst)


@task
def simple(c, src, output):
    slides(c, Path(src), Path(output), framework="simple")


@task
def impressjs(c, src, output):
    slides(c, Path(src), Path(output), framework="impressjs")


@task
def revealjs(c, src, output):
    slides(c, Path(src), Path(output), framework="revealjs")



# Copyright (C) 2022 H. Turgut Uyar <uyar@tekir.org>
#
# Kırlent is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Kırlent is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Kırlent.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path

from invoke import task

from . import slides
from .utils import MKDIR, relative_path, up_to_date


BUILD_PDF: str = "decktape reveal --size %(size)s %(in)s %(out)s"


@task
def decktape(c, src, output):
    src_path, output_path = Path(src), Path(output)
    slides.slides(c, src=src_path, output=output_path, framework="revealjs")
    slides_path = Path(c.config["revealjs:output"])
    dst_name = slides_path.name.replace("revealjs", "decktape")
    dst_path = Path(output, dst_name).with_suffix(".pdf")
    if not up_to_date(dst_path, [slides_path]):
        if not dst_path.parent.exists():
            c.run(MKDIR % {"dir": relative_path(dst_path.parent)})
        c.run(BUILD_PDF % {
            "size": c.slides.size,
            "in": relative_path(slides_path),
            "out": relative_path(dst_path),
        })



# Copyright (C) 2022 H. Turgut Uyar <uyar@tekir.org>
#
# Kırlent is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Kırlent is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Kırlent.  If not, see <http://www.gnu.org/licenses/>.

from contextlib import redirect_stdout
from html import escape as html_escape
from html.parser import HTMLParser
from io import StringIO
from itertools import dropwhile, zip_longest
from pathlib import Path
from typing import List, Sequence, Tuple, Union
from urllib.parse import urlparse
from xml.etree import ElementTree

from invoke import Context


MKDIR: str = "mkdir -p %(dir)s"
COPY: str = "cp %(src)s %(dst)s"


def newer(x: Path, y: Path) -> bool:
    return x.stat().st_mtime_ns > y.stat().st_mtime_ns


def up_to_date(target: Path, deps: Sequence[Path]) -> bool:
    return target.exists() and all(newer(target, dep) for dep in deps)


def relative_path(path: Path, wrt: Union[Path, None] = None) -> Path:
    start = wrt if wrt is not None else Path()
    parts = zip_longest(start.resolve().parts, path.resolve().parts)
    path_diff = dropwhile(lambda ps: ps[0] == ps[1], parts)
    up_parts, down_parts = zip(*path_diff)
    up_path = Path(*[".." for p in up_parts if p is not None])
    down_path = Path(*[p for p in down_parts if p is not None])
    return Path(up_path, down_path)


class HTMLNormalizer(HTMLParser):
    VOID_ELEMENTS = frozenset({"br", "col", "hr", "img", "link", "meta"})

    def handle_starttag(self, tag, attrs):
        attribs = []
        for attr_name, attr_value in attrs:
            if attr_value is None:
                attr_value = ""
            markup = '%(name)s="%(value)s"' % {
                "name": attr_name,
                "value": html_escape(attr_value, quote=True),
            }
            attribs.append(markup)
        line = "<%(tag)s%(attrs)s%(slash)s>" % {
            "tag": tag,
            "attrs": (" " + " ".join(attribs)) if len(attribs) > 0 else "",
            "slash": "/" if tag in HTMLNormalizer.VOID_ELEMENTS else "",
        }
        print(line, end="")

    def handle_endtag(self, tag):
        print(f"</{tag}>", end="")

    def handle_startendtag(self, tag, attrs):
        self.handle_starttag(tag, attrs)

    def handle_data(self, data):
        print(html_escape(data), end="")

    def error(self, message):
        """Ignore errors."""


def html_to_xhtml(document: str) -> str:
    out = StringIO()
    normalizer = HTMLNormalizer()
    with redirect_stdout(out):
        normalizer.feed(document)
    return out.getvalue()


def refs(content: str) -> List[Tuple[str, str]]:
    root = ElementTree.fromstring(html_to_xhtml(content))
    return [
        (ref.tag, ref.get(attr, ""))
        for tag, attr in (("link", "href"), ("script", "src"), ("img", "src"))
        for ref in root.findall(f".//{tag}[@{attr}]")
    ]


def collect_assets(c: Context, content: str, base: Path, wrt: Path) -> str:
    content_ = content
    for tag, ref in refs(content):
        url = urlparse(ref)
        if url.scheme not in {"", "file"}:
            continue

        asset_src = Path(url.path) if tag == "link" else Path(wrt, url.path)
        asset_dst = Path(base, asset_src.name)
        if not up_to_date(asset_dst, [asset_src]):
            c.run(COPY % {
                "src": relative_path(asset_src),
                "dst": relative_path(asset_dst),
            })
        pos = ref.rfind(asset_dst.name)
        content_ = content_.replace(ref, ref[pos:])
    return content_
